/*

	ent-ghost
	Copyright [2011-2013] [Jack Lu1]

	This file is part of the ent-ghost source code.

	ent-ghost is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	ent-ghost source code is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with ent-ghost source code. If not, see <http://www.gnu.org/licenses/>.

	ent-ghost is modified from GHost++ (http://ghostplusplus.googlecode.com/)
	GHost++ is Copyright [2008] [Trevor Hogan]

*/

#include "ghost.h"
#include "util.h"
#include "socket.h"
#include "commandpacket.h"
#include "keyprotocol.h"
#include "keyclient.h"

//
// CKeyClient
//

CKeyClient :: CKeyClient( string nServer, uint16_t nPort, string nRealmServer, bool nTFT, string nToken ) : m_Connected( false ), m_Server( nServer ), m_Port( nPort ), m_RealmServer( nRealmServer ), m_TFT( nTFT ), m_Token( nToken ), m_LastNullTime( 0 ), m_LastDisconnectedTime( 0 ), m_LastRequestTime( 0 )
{
	m_Socket = new CTCPClient( );
	m_Protocol = new CKeyProtocol( );
}

CKeyClient :: ~CKeyClient( )
{
	delete m_Socket;
	delete m_Protocol;

	while( !m_Packets.empty( ) )
	{
		delete m_Packets.front( );
		m_Packets.pop( );
	}
}

void CKeyClient :: MakeRequest( )
{
	Reset( );
	
	if( m_TFT )
		m_OutPackets.push( m_Protocol->SEND_KEY_OBTAIN( m_RealmServer, "tft" ) );
	else
		m_OutPackets.push( m_Protocol->SEND_KEY_OBTAIN( m_RealmServer, "roc" ) );

	m_LastRequestTime = GetTime( );
}

bool CKeyClient :: IsReady( )
{
	return !m_CDKeyROC.empty( ) && ( !m_TFT || !m_CDKeyTFT.empty( ) );
}

void CKeyClient :: Reset( )
{
	m_LastRequestTime = 0;
	m_CDKeyROC = "";
	m_CDKeyTFT = "";
}

unsigned int CKeyClient :: SetFD( void *fd, void *send_fd, int *nfds )
{
	if( !m_Socket->HasError( ) && m_Socket->GetConnected( ) )
	{
		m_Socket->SetFD( (fd_set *)fd, (fd_set *)send_fd, nfds );
		return 1;
	}

	return 0;
}

void CKeyClient :: Update( void *fd, void *send_fd )
{
	if( m_LastRequestTime != 0 && !IsReady( ) && GetTime( ) - m_LastRequestTime > 5 )
	{
		m_CDKeyROC = "FFFFFFFFFFFFFFFFFFFFFFFFFF";
		m_CDKeyTFT = "FFFFFFFFFFFFFFFFFFFFFFFFFF";
		m_LastRequestTime = 0;
	}

	if( m_Socket->HasError( ) )
	{
		CONSOLE_Print( "[Key " + m_RealmServer + ":" + m_Server + ":" + UTIL_ToString( m_Port ) + "] disconnected from key server due to socket error" );
		m_Socket->Reset( );
		m_LastDisconnectedTime = GetTime( );
		m_Connected = false;
		return;
	}

	if( !m_Socket->GetConnecting( ) && !m_Socket->GetConnected( ) && m_Connected )
	{
		CONSOLE_Print( "[Key " + m_RealmServer + ":" + m_Server + ":" + UTIL_ToString( m_Port ) + "] disconnected from key server" );
		m_Socket->Reset( );
		m_LastDisconnectedTime = GetTime( );
		m_Connected = false;
		return;
	}

	if( m_Socket->GetConnected( ) )
	{
		m_Socket->DoRecv( (fd_set *)fd );
		ExtractPackets( );
		ProcessPackets( );

		if( GetTime( ) - m_LastNullTime >= 50 )
		{
			m_Socket->PutBytes( m_Protocol->SEND_KEY_NULL( ) );
			m_LastNullTime = GetTime( );
		}

		while( !m_OutPackets.empty( ) )
		{
			m_Socket->PutBytes( m_OutPackets.front( ) );
			m_OutPackets.pop( );
		}

		m_Socket->DoSend( (fd_set *)send_fd );
		return;
	}

	if( m_Socket->GetConnecting( ) )
	{
		if( m_Socket->CheckConnect( ) )
		{
			// the connection attempt completed
			CONSOLE_Print( "[Key " + m_RealmServer + ":" + m_Server + ":" + UTIL_ToString( m_Port ) + "] connected" );
			m_Socket->PutBytes( m_Protocol->SEND_KEY_AUTH( m_Token ) );
			m_Socket->DoSend( (fd_set *)send_fd );
			m_Connected = true;
			m_LastNullTime = GetTime( );
			return;
		}
		else if( GetTime( ) - m_LastConnectionAttemptTime >= 15 )
		{
			// the connection attempt timed out (15 seconds)
			CONSOLE_Print( "[Key " + m_RealmServer + ":" + m_Server + ":" + UTIL_ToString( m_Port ) + "] connect timed out" );
			m_Socket->Reset( );
			m_LastDisconnectedTime = GetTime( );
			m_Connected = false;
			return;
		}
	}

	if( !m_Socket->GetConnecting( ) && !m_Socket->GetConnected( ) && !m_Connected && GetTime( ) >= m_LastDisconnectedTime + 15 )
	{
		CONSOLE_Print( "[Key " + m_RealmServer + ":" + m_Server + ":" + UTIL_ToString( m_Port ) + "] connecting to server [" + m_Server + "] on port " + UTIL_ToString( m_Port ) );
		m_Socket->Connect( string( ), m_Server, m_Port );
		m_LastConnectionAttemptTime = GetTime( );
		return;
	}

	return;
}

void CKeyClient :: ExtractPackets( )
{
	string *RecvBuffer = m_Socket->GetBytes( );
	BYTEARRAY Bytes = UTIL_CreateByteArray( (unsigned char *)RecvBuffer->c_str( ), RecvBuffer->size( ) );

	while( Bytes.size( ) >= 3 )
	{
		uint16_t Length = UTIL_ByteArrayToUInt16( Bytes, true );

		if( Length >= 3 )
		{
			if( Bytes.size( ) >= Length )
			{
				m_Packets.push( new CCommandPacket( 0, Bytes[2], BYTEARRAY( Bytes.begin( ), Bytes.begin( ) + Length ) ) );
				*RecvBuffer = RecvBuffer->substr( Length );
				Bytes = BYTEARRAY( Bytes.begin( ) + Length, Bytes.end( ) );
			}
			else
				return;
		}
		else
		{
			CONSOLE_Print( "[Key " + m_RealmServer + ":" + m_Server + ":" + UTIL_ToString( m_Port ) + "] error - received invalid packet from key server (bad length), disconnecting" );
			m_Socket->Disconnect( );
			return;
		}
	}
}

void CKeyClient :: ProcessPackets( )
{
	while( !m_Packets.empty( ) )
	{
		CCommandPacket *Packet = m_Packets.front( );
		m_Packets.pop( );

		if( Packet->GetID( ) == CKeyProtocol :: KEY_OBTAIN )
		{
			CKeyInfo *keyInfo = m_Protocol->RECEIVE_KEY_OBTAIN( Packet->GetData( ) );

			if( keyInfo )
			{
				CONSOLE_Print( "[Key " + m_RealmServer + ":" + m_Server + ":" + UTIL_ToString( m_Port ) + "] debug: received keys (" + keyInfo->GetType( ) + ")" );

				if( keyInfo->GetType( ) == "roc" )
					m_CDKeyROC = keyInfo->GetKey( );
				else if( keyInfo->GetType( ) == "tft" )
					m_CDKeyTFT = keyInfo->GetKey( );
			}
			else
				CONSOLE_Print( "[Key " + m_RealmServer + ":" + m_Server + ":" + UTIL_ToString( m_Port ) + "] debug: received bad key packet" );
		}
		else
		{
			CONSOLE_Print( "[Key " + m_RealmServer + ":" + m_Server + ":" + UTIL_ToString( m_Port ) + "] debug: received bad packet type " + UTIL_ToString( Packet->GetID( ) ) );
		}

		delete Packet;
	}
}
